import unittest

from zope.testing import doctest
from zope.testing import doctestunit
from zope.component import testing

OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


def test_suite():
    return unittest.TestSuite([

        doctestunit.DocTestSuite(
            module='xm.booking.browser.add',
            setUp=testing.setUp, tearDown=testing.tearDown),

        doctestunit.DocFileSuite(
            'tests.txt',
            optionflags=OPTIONFLAGS,
            package='xm.booking.timing'),

        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from Acquisition import aq_inner
from kss.core import kssaction
from plone.app.kss.plonekssview import PloneKSSView
from Products.Five.browser import BrowserView
from viewlets import Bookings as BookingsViewlet
from viewlets import BookingForm as BookingFormViewlet


def create_booking(context, title='Booking', hours=0, minutes=0):
    """Create a booking.

    We introduce a Mock Booking class for testing.

    >>> class MockBooking(object):
    ...     def __init__(self, **kwargs):
    ...         for key, value in kwargs.items():
    ...             self.__setattr__(key, value)

    Let's try this.

    >>> booking1 = MockBooking(id=42, blah='h2g2')
    >>> booking1.id
    42
    >>> booking1.blah
    'h2g2'
    >>> booking1.title
    Traceback (most recent call last):
    ...
    AttributeError: 'MockBooking' object has no attribute 'title'


    For now Bookings are added in Tasks, so we create a Mock Task
    class.

    >>> class MockTask(object):
    ...     def __init__(self):
    ...         self.items = []
    ...     def invokeFactory(self, type, **kwargs):
    ...         if type != 'Booking':
    ...             raise Exception('We want only Bookings.')
    ...         self.items.append(MockBooking(**kwargs))
    ...     def objectIds(self):
    ...         return [x.id for x in self.items]
    >>> context = MockTask()
    >>> context.objectIds()
    []
    >>> context.invokeFactory('Solfatara')
    Traceback (most recent call last):
    ...
    Exception: We want only Bookings.


    XXX We may want to put these Mocks and their tests into another
    class that we can use in other places as well.

    Try a few stupid things.

    >>> create_booking()
    Traceback (most recent call last):
    ...
    TypeError: create_booking() takes at least 1 argument (0 given)
    >>> create_booking(context, id='Peroni')
    Traceback (most recent call last):
    ...
    TypeError: create_booking() got an unexpected keyword argument 'id'

    Right, the basics seem to work.  Now we go and create some
    Bookings with this function.  We have some defaults in place.

    >>> create_booking(context)
    >>> context.objectIds()
    ['1']
    >>> booking = context.items[0]
    >>> booking.title
    'Booking'
    >>> booking.hours
    0
    >>> booking.minutes
    0

    Now add some non default values.

    >>> create_booking(context, title='Buongiorno', hours=3, minutes=15)
    >>> context.objectIds()
    ['1', '2']
    >>> booking = context.items[-1]
    >>> booking.title
    'Buongiorno'
    >>> booking.hours
    3
    >>> booking.minutes
    15

    """
    idx = 1
    while str(idx) in context.objectIds():
        idx = idx + 1
    context.invokeFactory('Booking', id=str(idx), title=title,
                          hours=hours, minutes=minutes)


class Create(BrowserView):

    def __call__(self):
        form = self.request.form
        title = form.get('title', '')
        hours = form.get('hours', 0)
        minutes = form.get('minutes', 0)
        context = aq_inner(self.context)
        create_booking(self.context, title=title, hours=hours, minutes=minutes)
        self.request.response.redirect(self.context.absolute_url())


class Add(PloneKSSView):

    @kssaction
    def add_booking(self):
        # We *really* need the inner acquisition chain for context
        # here.  Otherwise the aq_parent is the view instead of the
        # story, which means the totals for the story are not
        # recalculated.  Sneaky! :)
        context = aq_inner(self.context)
        create_booking(context, title=self.request.form.get('title'),
                       hours=self.request.get('hours'),
                       minutes=self.request.get('minutes'))
        core = self.getCommandSet('core')

        # Refresh the booking table
        viewlet = BookingsViewlet(context, self.request, None, None)
        viewlet.update()
        rendered = viewlet.render()
        selector = core.getHtmlIdSelector('booking-table')
        core.replaceHTML(selector, rendered)

        # Refresh the add booking form
        viewlet = BookingFormViewlet(context, self.request, None, None)
        viewlet.update()
        rendered = viewlet.render()
        selector = core.getHtmlIdSelector('add-booking')
        core.replaceHTML(selector, rendered)



import logging

from zope.interface import implements
from zope.annotation.interfaces import IAnnotations
from persistent.dict import PersistentDict
from Products.CMFPlone import CatalogTool as catalogtool
from interfaces import IActualHours
from interfaces import IEstimate
from interfaces import ISizeEstimate


logger = logging.getLogger('bookings')


class ActualHoursContainer(object):
    """An adapter for actual hours containers, like Task.
    """

    implements(IActualHours)
    ANNO_KEY = 'extrememanagement.actual'

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(context)
        self.metadata = annotations.get(self.ANNO_KEY, None)
        if self.metadata is None:
            annotations[self.ANNO_KEY] = PersistentDict()
            self.metadata = annotations[self.ANNO_KEY]

    def __get_actual_time(self):
        return self.metadata.get('actual_time', 0.0)
    def __set_actual_time(self, v):
        if not isinstance(v, float):
            v = float(v)
        self.metadata['actual_time'] = v
    actual_time = property(__get_actual_time, __set_actual_time)

    def recalc(self):
        """Recalculate the total booked hours for this container.
        """
        context = self.context
        #logger.info("XXXRecalcing for %s", context)
        total = 0.0
        for obj in context.contentValues():
            #logger.info("XXXWaking up %s", obj)
            actual = IActualHours(obj, None)
            if actual is not None:
                total += actual.actual_time
        self.actual_time = total
        context.setModificationDate()
        context.reindexObject(idxs=['actual_time', 'modified',
                                    'ModificationDate'])


class EstimateContainer(ActualHoursContainer):
    implements(IEstimate)
    ANNO_KEY = 'extrememanagement.estimate'

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(context)
        self.metadata = annotations.get(self.ANNO_KEY, None)
        if self.metadata is None:
            annotations[self.ANNO_KEY] = PersistentDict()
            self.metadata = annotations[self.ANNO_KEY]

    def __get_estimate(self):
        return self.metadata.get('estimate', 0.0)
    def __set_estimate(self, v):
        if not isinstance(v, float):
            v = float(v)
        self.metadata['estimate'] = v
    estimate = property(__get_estimate, __set_estimate)

    def recalc(self):
        """Recalculate the total of estimates for this container.
        """
        context = self.context
        total = 0.0
        for obj in context.contentValues():
            adapted = IEstimate(obj, None)
            if adapted is not None:
                total += adapted.estimate
        self.estimate = total
        context.setModificationDate()
        context.reindexObject(idxs=['estimate', 'modified',
                                    'ModificationDate'])


def actual(object, portal, **kw):
    adapted = IActualHours(object, None)
    if adapted is not None:
        return adapted.actual_time
    return None


def duration_estimate(object, portal, **kw):
    adapted = IEstimate(object, None)
    if adapted is not None:
        return adapted.estimate
    return None


def size_estimate(object, portal, **kw):
    adapted = ISizeEstimate(object, None)
    if adapted is not None:
        return adapted.size_estimate
    return None


catalogtool.registerIndexableAttribute('actual_time', actual)
catalogtool.registerIndexableAttribute('estimate', duration_estimate)
catalogtool.registerIndexableAttribute('size_estimate', size_estimate)



from Acquisition import aq_inner
from plone.app.layout.viewlets import ViewletBase
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from zope.interface import Interface
from zope.interface import implements


class BookingFormInterface(Interface):

    def allowed():
        """Is the user allowed to add a booking here?
        """
        pass


class Bookings(ViewletBase):

    render = ViewPageTemplateFile("bookingtable.pt")


class BookingForm(ViewletBase):
    implements(BookingFormInterface)

    def allowed(self):
        """Is the user allowed to add a booking here?
        """
        context = aq_inner(self.context)
        mem = getToolByName(context, 'portal_membership')
        return mem.checkPermission('eXtremeManagement: Add Booking', context)

    render = ViewPageTemplateFile("add.pt")



from Acquisition import aq_inner, aq_parent

from interfaces import IActualHours
from interfaces import IEstimate


def recursivelyRecalc(object, adapter):
    current = aq_inner(object)
    anno = adapter(current, None)
    while anno is not None:
        anno.recalc()
        current = aq_parent(current)
        anno = adapter(current, None)


def addedActualHours(object, event):
    """An object with actual hours (a Booking or a parent usually) was
    added, moved or removed.  So we reindex the old and new
    containers, if they exist.

    Some notes.

    Adding Bookings
    ---------------

    When *adding* a Booking in the browser the IObjectMovedEvent is
    fired more often than you would care to know.  But it is important
    to understand this anyway, so we had better care.

    All events are IObjectAddedEvents, except for the last one, which
    is an IObjectMovedEvent.  Note though that IObjectAddedEvents
    inherits from IObjectMovedEvent.

    1. The Booking is first created in the portal factory.

      - event.oldParent is None.

      - event.newParent is then a TempFolder

    2. This event is called a second time with those settings.  After
       this, you can fill in values in your browser.

    3. Then in the browser you click save so the event is fired again.

      - event.oldParent is None.

      - event.NewParent is still a TempFolder

    4. Then the event is fired another time:

      - event.oldParent is None.

      - event.newParent is a Task (unless you allow Bookings to be
        added in other content types)

      - nothing has been filled in yet, specifically the Booking has
        no hours or minutes yet

    5. Then the final step:

      - event.oldParent is a Task.

      - event.newParent is that same Task (!).  Ah, Martijn Pieters is
        right in saying that this is because the Booking is renamed.
        Makes sense.

      - So this is an IObjectMovedEvent.

      - the Booking now has values.

    This last time is the only time that we actually want to do
    something in this event handler.

    Actually, it looks like even that last time we do not have to do
    anything.  The Booking with hours and minutes is already in the
    Task and is just renamed.  At this point the total in the Task has
    already been recalculated, courtesy of the event handler that
    listens to IObjectModifiedEvents.


    When moving a Booking:
    ----------------------

    - oldParent is the old parent Task

    - newParent is the new parent Task


    When removing a Booking:
    ------------------------

    - oldParent is the old parent Task

    - newParent is None


    Moving a parent
    ---------------

    One more gotcha: when moving a Task with a child Booking to a new
    location, the parent of the Booking will remain that same Task, so
    no event really has to be handled here.  But the event of the Task
    moving to a new parent is actually dispatched to all its childs.
    In that case the object is a Booking and event.newParent is a
    Story, which seems strange.  In that case you need to be aware
    that event.object is that Task.

    So using an event a Booking is told that his parent has moved to a
    newParent.  In this event handler we do not have to do anything
    then.

    See zope.app.container.contained.py:dispatchToSublocations() which
    says: This suggests that location event handlers need to be aware
    that the objects they are called on and the event objects could be
    different.

    """
    if object != event.object:
        # The parent itself has moved and we do not care.
        return
    if event.oldParent == event.newParent:
        # Just a rename, so never mind
        return
    anno = IActualHours(object)
    if anno.actual_time == 0.0:
        # Nothing needs to be done here.
        return
    for parent in (event.oldParent, event.newParent):
        recursivelyRecalc(parent, IActualHours)


def modifiedActualHours(object, event):
    """An object with actual hours (a Booking or a parent usually) was
    modified.  So we reindex its parent too.
    """
    recursivelyRecalc(object, IActualHours)


def addedEstimate(object, event):
    """An object with an estimate (a Task or its parent usually) was
    added, moved or removed.  So we reindex the old and new
    containers, if they exist.
    """
    if object != event.object:
        # The parent itself has moved and we do not care.
        return
    if event.oldParent == event.newParent:
        # Just a rename, so never mind
        return
    anno = IEstimate(object)
    if anno.estimate == 0.0:
        # Nothing needs to be done here.
        return
    for parent in (event.oldParent, event.newParent):
        recursivelyRecalc(parent, IEstimate)


def modifiedEstimate(object, event):
    """An object with estimates (a Task or its parent usually) was
    modified.  So we reindex its parent too.
    """
    recursivelyRecalc(object, IEstimate)



from zope.interface import Interface
from zope.interface import Attribute
from zope.annotation.interfaces import IAttributeAnnotatable


class IActualHoursContainer(IAttributeAnnotatable):
    """Marker interface: this object has children with actual hours
    """

    def contentValues():
        """List the content objects in this container
        """


class IActualHours(Interface):
    """Actual hours and minutes worked
    """

    actual_time = Attribute("Actual time")

    def recalc():
        """Recalculate the total of bookings/actual hours.
        """


class IEstimateContainer(IAttributeAnnotatable):
    """Marker interface: this object has children with estimates
    """

    def contentValues():
        """List the content objects in this container
        """


class IEstimate(Interface):
    """Estimated time to work
    """

    estimate = Attribute("Estimate")

    def recalc():
        """Recalculate the total of estimates
        """


class ISizeEstimate(Interface):
    """Estimated size of work load

    Think Story Points or Ideal Days
    """

    size_estimate = Attribute("Size Estimate")

    def recalc():
        """Recalculate the total of size estimates
        """



from setuptools import setup, find_packages
import os.path

versionfile = open(os.path.join('xm', 'booking', 'version.txt'))
version = versionfile.read().strip()
versionfile.close()

readmefile = open(os.path.join('xm', 'booking', 'README.txt'))
readme = readmefile.read().strip()
readmefile.close()


setup(name='xm.booking',
      version=version,
      description="Bookings for eXtremeManagement",
      long_description=readme,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='eXtremeManagement booking',
      author='Maurits van Rees',
      author_email='m.van.rees@zestsoftware.nl',
      url="http://plone.org/products/extreme-management-tool/",
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['xm'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
